/*
 * Siemens ISAC / HSCX chipset driver for leased lines
 * ---------------------------------------------------
 * Version 0.51
 *
 * assembled/patched by Michael Hipp
 * copyright (c) 1996 by Michael Hipp
 * linux/drivers/isdn/teles/card.c: copyright (c) 1995,1996 Jan den Ouden
 * linux/drivers/net/skeleton.c:    copyright (c) 1992-1994 Donald Becker
 *
 * ---------------------------------------------------------------------
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (Version 2.0) as
 * published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ---------------------------------------------------------------------
 *
 * (probably) supported cards:
 *    Teles 8, 16.0, 16.2, 16.3 
 *    Creatix 16.0 , 16.3 
 *    Niccy 1016 
 * I've "tested" (loopback mode only) the driver with a Teles 16.0 
 *
 * I'm not maintaining this driver .. it was just an initial version
 * Anyone using this driver should do the maintenance
 *
 * The driver currently ignores all D-channel data.
 * This driver is _NO_ low level device for the isdn4linux package.
 * The driver defines his own network device (name: 'leased0')
 * You have to configure the device with:
 *     'ifconfig leased0 <myIP> pointopoint <peerIP>'
 * You need a scenario like:
 *     Machine1 --S0-Bus--|NT|------- [TELEKOM] ------|NT|--S0-Bus-- Machine2
 * to use the driver.
 * Currently the driver can only send and receive IP packets.
 * There is not ethernet bridging.
 *
 * compile with:
 *    gcc -O2 -Wall -m486 -fomit-frame-pointer -DMODULE -D__KERNEL__ -c isachscx.c
 *
 * load (a teles 16.0) with e.g:
 *   insmod isachscx io=0xd80 irq=12 mem=0xCC000 channel=2
 * load (a teles 16.3) with e.g:
 *   insmod isachscx io=0xd80 irq=12 mem=0x0 channel=2
 * or for loopback testmode (here 16.0) e.g:
 *   insmod isachscx io=0xd80 irq=12 mem=0xCC000 channel=3 loopback=1
 *
 * symbols:
 *   io:      set IO base address (set this to 0x0 for Teles8) .. (default: 0xd80)
 *   irq:     interrupt number (default: IRQ 12)
 *   mem:     shared memory base (set this to 0x0 for Teles 16.3) (default: 0xCC000)
 *   channel: time slot to use: (default: 1)
 *              1= use first timeslot
 *              2= use second timeslot
 *              3= use both timeslots in ONE stream (=one 16 bit timeslot)
 *                  maybe works only with a isachscx driver on both
 *                  ends of the link! It also often happens, that
 *                  timeslot 1 and timeslot 2 are out of sync.
 *              8= use both timeslots in TWO streams (=two 8 bit timeslots)
 *                  use this mode for 'out of sync' links ..
 *                  (experimental and not heavily tested!!)
 *   debug:   0=no debug, 1: a few msgs., 2: fifo-buffer logging
 *   devname: alternate network device name (default "leased0")
 *
 * Documentation:
 *   Siemens ISDN Subscriber Access Controller, ISAC S, PEB 2085,PEB 2086 
 *                (PEB 2186 (IOM-2 only) on newer cards)
 *   Siemens HSCX High-Level Serial Communications Controller Extended
 *                SAB 82525,SAB 82526 (PSB 21525)
 *   Address:
 *     Siemens AG
 *     LZF - Semiconductor Book Shop
 *     90713 Fuerth-Bislohe
 *     Germany
 *
 * known bugs:
 *    HSCX-A handling near untested 
 *    Never tested with a real leased line myself!! (but there ARE working solutions)
 */

/*
 * used sources:
 *   card.c from the isdn4linux package by Jan den Ouden
 *   skeleton.c from Linux net package by Donald Becker
 */

/*
 * changelog:
 *
 * 12.August.1996: some minor changes ..
 * 11.August.1996: added inital HSCX-A handling
 * 10.August.1996: some changes in the header information (MH)
 *                 added L1 activation code
 * 14.Mai.1996: inital version (MH)
 */

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/delay.h>

#include "ih_card.h"

#define byteout(addr,val) outb_p(val,addr)
#define bytein(addr) inb_p(addr)

#define byte unsigned char

extern int arp_find(unsigned char *, u32, struct device *, u32,
		    struct sk_buff *);


static const char* cardname = "isachscx";

#define IH_FLIPBUFLEN 2048

static int channel=0x1;
static int encap = 0x0;	/* 0=IP , (1=Ether, but don't use!!) */
static int chanbits=15;	/* bits-1 for mode channel=3 */

static int loopback=0; /* 1=test !!, 0=normal operation */
static int debug=0;
static int activate=1;

static int amask=0xff,bmask=0xff;

#define NUM_CHANNELS 2
#define NUM_BUFFERS  4

struct priv {
	struct enet_statistics stats;
	unsigned char flipbuf[NUM_CHANNELS][IH_FLIPBUFLEN];
	int buflen[NUM_CHANNELS];
	int encap;
	int channels[NUM_CHANNELS];
	struct sk_buff *skb[NUM_CHANNELS];
	int ph_state;
	int debug;
	int xfw[NUM_CHANNELS];
	int activate;
	struct sk_buff *first;
	struct sk_buff *last;
	int numbuffs;
	struct isachscx ih;
};

extern int ih_probe(struct device *dev);
static int ih_open(struct device *dev);
static int ih_send_packet(struct sk_buff *skb, struct device *dev);
static void ih_interrupt(int irq, void *dev_id, struct pt_regs *regs);
static int ih_close(struct device *dev);
static struct enet_statistics *ih_get_stats(struct device *dev);
static void ih_init(struct device *dev, int m);
static int ih_header(struct sk_buff *skb, struct device *dev, unsigned short type,
		void *daddr, void *saddr, unsigned plen);
static int ih_rebuild_header(void *buff, struct device *dev,
		ulong dst,struct sk_buff *skb);

int ih_probe(struct device *dev)
{
	struct priv *p;
	int ret,i;

	dev->priv = kmalloc( sizeof(struct priv) , GFP_KERNEL);
	p = (struct priv *) dev->priv;
	if(!p)
		return -ENOMEM;
	memset(dev->priv, 0, sizeof(struct priv));

	ret = ih_checkcard(&p->ih,dev->base_addr,(void *)dev->mem_start,dev->irq);

	if(ret < 0) {
		kfree(dev->priv);
		return -EAGAIN;
	}

	dev->base_addr = p->ih.iobase;
	dev->mem_start = (unsigned long) p->ih.membase;
	dev->irq = p->ih.irq;

	p->encap = encap;
	for(i=0;i<NUM_CHANNELS;i++) {
		p->skb[i] = NULL;
		p->xfw[i] = 1;
		p->buflen[i] = 0;
		p->channels[i] = 0;
	}
	p->numbuffs = 0;
	p->first = p->last = NULL;

	switch(channel) {
		case 0:
		case 4:
		default:
			p->channels[0] = p->channels[1] = 0;
			break;
		case 1:
		case 2:
		case 3:
		case 5:
		case 6:
		case 7:
			p->channels[1] = channel & 0x3;
			break;
		case 8:
			p->channels[1] = 1;
			p->channels[0] = 2;
			break;
	}

	p->debug = debug;
	p->ph_state = 0;
	p->activate = activate;

	/* Grab the region so that no one else tries to probe our ioports. */

	if(dev->base_addr)
		request_region(dev->base_addr, 8 , cardname);
	else {
		if (request_irq(dev->irq, &ih_interrupt, 0, cardname, NULL)) {
			return -EAGAIN;
		}	
	}

	dev->open		= ih_open;
	dev->stop		= ih_close;
	dev->hard_start_xmit = ih_send_packet;
	dev->get_stats	= ih_get_stats;

	dev->hard_header      = ih_header;
	dev->rebuild_header   = ih_rebuild_header;
	dev->hard_header_len  = 0;

	/* device INFO */
	dev->mtu              = 1500;
	dev->addr_len         = 0;
	dev->type             = ARPHRD_SLIP; /* dummy */

	for (i = 0; i < DEV_NUMBUFFS; i++)  {
		skb_queue_head_init(&dev->buffs[i]);
	}

	if(loopback)
		dev->flags = IFF_LOOPBACK;
	else
		dev->flags = IFF_POINTOPOINT;

	dev->family     = AF_INET;
	dev->pa_addr    = 0;
	dev->pa_brdaddr = 0;
	dev->pa_mask    = 0;
	dev->pa_alen    = 4;

	return 0;
}

static int ih_header(struct sk_buff *skb, struct device *dev, unsigned short type,
		void *daddr, void *saddr, unsigned plen)
{
	struct priv *p = (struct priv *) dev->priv;

	if(p->encap == 1)
		return 0; /* eth_header(skb,dev,type,daddr,saddr,plen); */
	else
		return 0;

}


static int ih_rebuild_header(void *buff, struct device *dev,
		ulong dst,struct sk_buff *skb)
{
	int ret = 0;
	struct priv *p = (struct priv *) dev->priv;
	
	if(p->encap == 1) {
		struct ethhdr *eth = (struct ethhdr *)buff;
		
		if(eth->h_proto != htons(ETH_P_IP)) {
			printk(KERN_WARNING
			       "ih_header: Don't know how to resolve type %d addresses?\n",
			       (int)eth->h_proto);
			memcpy(eth->h_source, dev->dev_addr, dev->addr_len);
			return 0;
		}
#ifdef CONFIG_INET
		ret = arp_find((unsigned char *)&(eth->h_dest), dst, 
			dev, dev->pa_addr,skb)? 1 : 0;
#endif  
	}
	return ret;
}


static int ih_open(struct device *dev)
{
	struct priv *p = (struct priv *) dev->priv;

	if(dev->base_addr)
		if (request_irq(dev->irq, &ih_interrupt, 0, cardname, NULL)) {
			return -EAGAIN;
		}

	irq2dev_map[dev->irq] = dev;

	if(p->channels[0]) {
		amask = 0x00;
		bmask &= 0xfb;
	}
	if(p->channels[1]) {
		bmask &= 0x04;
	}

	ih_init(dev, 1);

	dev->tbusy = 0;
	dev->interrupt = 0;
	dev->start = 1;

	MOD_INC_USE_COUNT;

	return 0;
}

static void ih_enqueue_skb(struct priv *p,struct sk_buff *skb)
{
	long flags;

	save_flags(flags);
	cli();

	skb->next = NULL;
	if(p->last) {
		p->last->next = skb;
		p->last = skb;		
	} else {
		p->last = p->first = skb;
	}
	p->numbuffs++;

	restore_flags(flags);
}

static struct sk_buff *ih_dequeue_skb(struct priv *p)
{
	long flags;
	struct sk_buff *skb;

	save_flags(flags);
	cli();

	if(p->first) {
		skb = p->first;
		p->first = skb->next;
		skb->next = NULL;
		if(!p->first)
			p->last = NULL;
		p->numbuffs--;
	}
	else
		skb = NULL;

	restore_flags(flags);	
	return skb;
}


/*
 * Fill XFIFO buffers ...  doesn't check XFW bit!! 
 */

static void ih_fill_fifo(struct device *dev,int hscx,struct sk_buff *skb)
{
	int count;
	unsigned char *ptr;
	struct priv *p;
	p = (struct priv *) dev->priv;

	if ( (count = skb->len) <= 0)
		return;

	if (count > 32) {
		count = 32;
	}

	ptr = skb->data;
	skb_pull(skb,count);

#if 0
	{
	struct priv *p = (struct priv *) dev->priv;
		if(p->debug > 1)
		{
		  int i;
		  for(i=0;i<count;i++)
		    printk("%02x ",ptr[i]);
		  printk("\n");
		}
	}
#endif

	writehscx_s(&p->ih,hscx,ptr , count);
	writehscxCMDR(&p->ih,hscx, 0x8 );

	dev->trans_start = jiffies;
}

static void ih_empty_fifo(struct priv *p,int hscx,struct device *dev, int count)
{
	unsigned char *ptr;
	int cnt=count;

	if(p->buflen[hscx] + count > IH_FLIPBUFLEN) {
		printk(KERN_WARNING "isachscx: Can't hold data in buffer!\n");
		p->buflen[hscx] = 0;
	}

	ptr = p->flipbuf[hscx] + p->buflen[hscx];
	p->buflen[hscx] += count;

	readhscx_s(&p->ih, hscx,ptr, count);
	writehscxCMDR(&p->ih, hscx, 0x80);

	if(p->debug > 1)
	{
	  int i;
	  ptr = p->flipbuf[hscx] + p->buflen[hscx] - cnt;
	  for(i=0;i<cnt;i++)
	    printk("%02x ",ptr[i]);
	  printk("\n");
	}

}

static int ih_send_packet(struct sk_buff *skb, struct device *dev)
{
	struct priv *p = (struct priv *)dev->priv;

	if (dev->tbusy) {
		int tickssofar = jiffies - dev->trans_start;
		if (tickssofar < 20)
			return 1;
		printk(KERN_WARNING "%s: transmit timed out?\n", dev->name);
		ih_init(dev, 1);
		dev->tbusy=0;
		dev->trans_start = jiffies;
	}
	if (skb == NULL) {
		dev_tint(dev);
		return 0;
	}

	if (set_bit(0, (void*)&dev->tbusy)) {
		printk(KERN_WARNING "%s: Transmitter access conflict.\n", dev->name);
		return 1;
	}
	else {

		if(skb->len <= 0) {
			dev_kfree_skb(skb,FREE_WRITE);
			dev->tbusy = 0;
			return 0;
		}
		else {
			long flags;
			int i;

#ifdef ISACHSCX_VJ_COMPRESS
			skb = ih_encode_vj(skb);
#endif
#ifdef ISACHSCX_BSD_COMPRESS
			skb = ih_encode_bsd(skb);
#endif

			ih_enqueue_skb(p,skb);

			save_flags(flags);
			cli();
			for(i=0;i<NUM_CHANNELS;i++) {
				if(p->channels[i] && !p->skb[i] && !p->xfw[i]) {
					if( (p->skb[i] = ih_dequeue_skb(p)) )
						ih_fill_fifo(dev,i,p->skb[i]);
					break;
				}
			}
			dev->tbusy = (p->numbuffs >= NUM_BUFFERS);
			restore_flags(flags);
		}

	}
	return 0;
}

static void ih_ph_command(struct device *dev,unsigned int c)
{
	struct priv *p = (struct priv *) dev->priv;
	printk(KERN_DEBUG "ph_command %d\n", c);
	writeisac(&p->ih, ISAC_CIX0, (c << 2) | 3);
}

static void ih_new_ph(struct device *dev,int state)
{
	switch (state) {
		case (0):
		case (6):
			ih_ph_command(dev, 0);
			break;
		case (7):
			ih_ph_command(dev, 9);
			break;
		case (12):
			ih_ph_command(dev, 8);
			break;
		case (13):
			ih_ph_command(dev, 9);
			break;
		case (4):
		case (8):
			break;
		default:
			break;
	}
}


static void ih_hscx_interrupt(struct device *dev,struct priv *p,int hscx,int val,int exir)
{
	if (exir) {
		byte v1 = readhscx(&p->ih, hscx, HSCX_EXIR);
		printk(KERN_ERR "HSCX %c EXIR %x ",'A'+hscx,v1);
		if(v1 & 0x40) {
			printk(KERN_ERR "Xmit underflow\n");
			p->stats.tx_errors++;
			p->stats.tx_fifo_errors++;
			if (p->skb[hscx]) {
				dev_kfree_skb(p->skb[hscx],FREE_WRITE);
				p->skb[hscx] = NULL;
				dev->tbusy = 0;
				mark_bh(NET_BH);
			}
		}
		if(v1 & 0x10) {
			p->buflen[hscx] = 0;
			p->stats.rx_dropped++;
			writehscxCMDR(&p->ih, hscx, 0x80 );
		}
	}

	if (val & 0x80) {	/* RME */
		unsigned char r;
		int cnt = readhscx(&p->ih, hscx,HSCX_RBCL) & 0x1f;
		if(cnt == 0)	
			cnt = 32;
		ih_empty_fifo(p,hscx,dev,cnt);
		if(p->buflen[hscx] == 1)
			printk(KERN_ERR "%s: empty frame!\n",dev->name);
		r = p->flipbuf[hscx][p->buflen[hscx]-1];
		p->buflen[hscx]--;
		if (p->buflen[hscx] < 0 || (r & 0xf0) != 0xa0) {
			p->stats.rx_errors++;
			if (!r & 0x80) {
				printk(KERN_WARNING "isachscx: HSCX invalid frame\n");
				p->stats.rx_frame_errors++;
			}
			if (r & 0x40) {
				printk(KERN_WARNING "isachscx: HSCX RDO\n");
				p->stats.rx_fifo_errors++;
			}
			if (!r & 0x20) {
				printk(KERN_WARNING "isachscx: HSCX CRC error\n");
				p->stats.rx_crc_errors++;
			}
			writehscxCMDR(&p->ih, hscx,0x80);
			p->buflen[hscx] = 0;
		}
		else {
			struct sk_buff *skb;
			skb = dev_alloc_skb(p->buflen[hscx]);
			if (!skb)  {
				printk(KERN_WARNING "%s: memory squeeze, dropping packet.\n",dev->name);
				p->stats.rx_dropped++;
				p->buflen[hscx] = 0;
			}
			else {
				skb->dev = dev;
				memcpy(skb_put(skb,p->buflen[hscx]), p->flipbuf[hscx], p->buflen[hscx]);
#ifdef ISACHSCX_BSD_COMPRESS
				skb = ih_decode_bsd(dev,skb);
#endif
#ifdef ISACHSCX_VJ_COMPRESS
				skb = ih_decode_vj(dev,skb);
#endif
				if(p->encap == 1) {
					skb->protocol=eth_type_trans(skb,dev);
				}
				else {
					skb->protocol=htons(ETH_P_IP);
					skb->mac.raw=skb->data;
				}
				p->stats.rx_packets++;
				p->buflen[hscx] = 0;
				netif_rx(skb);
			}
		}
	}
	if (val & 0x40) {       /* RPF */
		ih_empty_fifo(p,hscx,dev,32);
	}
	if (val & 0x10) {	/* XPR */
		if (p->skb[hscx]) {
			p->xfw[hscx] = 0;	/* not necessary */
			if(p->skb[hscx]->len <= 0) {
				p->stats.tx_packets++;
				dev_kfree_skb(p->skb[hscx],FREE_WRITE);

				if( !(p->skb[hscx] = ih_dequeue_skb(p)) ) {
					p->xfw[hscx] = 1;
					writehscxCMDR(&p->ih, hscx, 0xa );
				}
				else {
					if(!( dev->tbusy = (p->numbuffs >= NUM_BUFFERS) ) )
						mark_bh(NET_BH);
					writehscxCMDR(&p->ih, hscx, 0x2 );
				}
			}
			if(p->skb[hscx])
				ih_fill_fifo(dev,hscx,p->skb[hscx]);
		}
		else {
			if( (p->skb[hscx] = ih_dequeue_skb(p)) ) 
				ih_fill_fifo(dev,hscx,p->skb[hscx]);
			p->xfw[hscx] = 0;
			if(!( dev->tbusy = (p->numbuffs >= NUM_BUFFERS) ) )
				mark_bh(NET_BH);
		}
	}
}

static void ih_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	struct device *dev = (struct device *)(irq2dev_map[irq]);
	struct priv *p;
	unsigned char val;

	if (dev == NULL) {
		if(debug > 1)
			printk(KERN_WARNING "%s: irq %d for unknown device.\n", cardname, irq);
		return;
	}
	dev->interrupt = 1;
	p = (struct priv *)dev->priv;

	val = readhscx(&p->ih, 1, HSCX_ISTA);

	if(val & 0xf9)
		ih_hscx_interrupt(dev,p,1,val,val & 0x1);
	if(val & 0x06) {
		unsigned char val1=0;
		if(val & 0x4)
			val1 = readhscx(&p->ih, 0, HSCX_ISTA);
		ih_hscx_interrupt(dev,p,0,val1,val & 0x2);
	}

	val = readisac(&p->ih, ISAC_ISTA);
	if (val & 0x04) {
		if(p->activate) {
			p->ph_state = (readisac(&p->ih, ISAC_CIX0) >> 2) & 0xf;
			printk(KERN_DEBUG "l1state %d\n", p->ph_state);
			ih_new_ph(dev,p->ph_state);
		}
	}

	writeisac(&p->ih, ISAC_MASK, 0xFF);
	writehscx(&p->ih, 0, HSCX_MASK, 0xFF);
	writehscx(&p->ih, 1, HSCX_MASK, 0xFF);
	if(p->activate)
		writeisac(&p->ih, ISAC_MASK, 0xFB);
	writehscx(&p->ih, 0, HSCX_MASK, amask);
	writehscx(&p->ih, 1, HSCX_MASK, bmask);

	dev->interrupt = 0;
	return;
}

static int ih_close(struct device *dev)
{
	struct priv *p = (struct priv *) dev->priv;
	struct sk_buff *skb;
	int i;

	dev->tbusy = 1;
	dev->start = 0;

	if(dev->base_addr)
		free_irq(dev->irq, NULL);
	irq2dev_map[dev->irq] = 0;

	for(i=0;i<NUM_CHANNELS;i++) {
		if(p->skb[i]) {
			dev_kfree_skb(p->skb[i],FREE_WRITE);
			p->skb[i] = NULL;
		}
	}
	while( (skb=ih_dequeue_skb(p)) ) {
		dev_kfree_skb(skb,FREE_WRITE);
	}

	amask = bmask = 0xff;

	ih_init(dev,0);

	MOD_DEC_USE_COUNT;

	return 0;

}

static void ih_init(struct device *dev, int mode1)
{
	struct priv *p = (struct priv *) dev->priv;
	int hscx,mode;
	int i;
	struct sk_buff *skb;

	while( (skb = ih_dequeue_skb(p)) ) {
		dev_kfree_skb(skb,FREE_WRITE);
	}

	for(i=0;i<NUM_CHANNELS;i++) {
		if(p->skb[i]) {
			dev_kfree_skb(p->skb[i],FREE_WRITE);
			p->skb[i] = NULL;
		}
		p->xfw[i] = 1;
		p->buflen[i] = 0;
	}

	if(mode1)
		mode = loopback ? 4 : 2;
	else
		mode = 0;

	writehscx(&p->ih, 0, HSCX_MASK, 0xFF);
	writehscx(&p->ih, 1, HSCX_MASK, 0xFF);
	writeisac(&p->ih, ISAC_MASK, 0xff);

	ih_initisac(&p->ih,0x0,0x0);

	for(hscx=0;hscx<2;hscx++) {
		if(p->channels[hscx] == 1 || p->channels[hscx] == 2)
			ih_hscxmode(&p->ih,hscx,mode,p->channels[hscx]-1,7);
		else if(p->channels[hscx])
			ih_hscxmode(&p->ih,hscx,mode,0,chanbits);
		else
			ih_hscxmode(&p->ih,hscx,0,0,7);
	}

	writeisac(&p->ih, ISAC_MASK, 0xFB);
	writehscx(&p->ih, 0, HSCX_MASK, amask);
	writehscx(&p->ih, 1, HSCX_MASK, bmask);
}

static struct enet_statistics *ih_get_stats(struct device *dev)
{
	struct priv *p = (struct priv *) dev->priv;
	return &p->stats;
}

#ifdef ISACHSCX_VJ_COMPRESS

void ih_init_vj(struct device *dev)
{
	struct priv *p = (struct priv *) dev->priv;
	p->slcomp = slhc_init(16, 16);
}

void ih_exit_vj(struct device *dev)
{
	struct priv *p = (struct priv *) dev->priv;

	slhc_free(p->slcomp);
	p->slcomp = NULL;
}

static struct skbuff *ih_encode_vj(struct device *dev,struct skbuff *skb)
{
	struct skbuff *skb;

	new_skb = dev_alloc_skb(skb->len);
	if(new_skb) {
		u_char *buf;
		int pktlen;
		struct priv *p = (struct priv *) dev->priv;

		new_skb->dev = skb->dev;
		new_skb->free = 1;
		skb_put(new_skb,skb->len);
		buf = skb->data;

		pktlen = slhc_compress(p->slcomp, skb->data, skb->len, new_skb->data, &buf, 0);

		if(buf != skb->data) {
			if(new_skb->data != buf)
				printk(KERN_ERR "%s: FATAL error after slhc_compress!!\n",dev->name);
			dev_kfree_skb(skb,FREE_WRITE);
			skb = new_skb;
		}
		else {
			dev_kfree_skb(new_skb,0 /* FREE_WRITE */ );
		}

		skb_trim(skb,pktlen);
	}

	return skb;
}

static struct skbuff *ih_decode_vj(struct device *dev,struct skbuff *skb)
{
	int pkt_len;
	struct priv *p = (struct priv *) dev->priv;

	if(skb->data[0] & SL_TYPE_COMPRESSED_TCP) {
		struct sk_buff *skb_old = skb;
		skb = dev_alloc_skb(skb_old->len + 40);
		skb_old->free = 1;

		if (!skb) {
			printk(KERN_WARNING "%s: Memory squeeze, dropping packet.\n", dev->name);
			p->stats.rx_dropped++;
			dev_kfree_skb(skb_old,0 /* FREE_READ */ );
			return NULL;
		}
		skb->dev = skb_old->dev;
		skb_put(skb,skb_old->len + 40);
		memcpy(skb->data, skb_old->data, skb_old->len);
		skb->mac.raw = skb->data;
		pkt_len = slhc_uncompress(p->slcomp, skb->data, skb_old->len);
		dev_kfree_skb(skb_old,0 /* FREE_READ */ );
		if(pkt_len < 0) {
			skb->free = 1;
			dev_kfree_skb(skb, 0 /* FREE_READ */ );
			p->stats.rx_dropped++;
			return NULL;
		}
		skb_trim(skb, pkt_len);
		skb->protocol = htons(ETH_P_IP);
	}
	else {
		if(skb->data[0] >= SL_TYPE_UNCOMPRESSED_TCP) {
			if(slhc_remember(p->slcomp, skb->data, skb->len) <= 0) {
				printk(KERN_WARNING "%s: received illegal VJC_UNCOMP frame!\n",dev->name);
				p->stats.rx_dropped++;
				skb->free = 1;
				dev_kfree_skb(skb,0 /* FREE_READ */ );
				return NULL;
			}
		}
	}
	return skb;
}

#endif

#ifdef ISACHSCX_BSD_COMPRESS

static struct skbuff *ih_encode_bsd(struct skbuff *skb)
{

}

static struct skbuff *ih_decode_bsd(struct skbuff *skb)
{

}

#endif

#ifdef MODULE

static char devicename[9] = "leased0";
static struct device ih_device = {
	devicename, 
	0, 0, 0, 0,
	0, 0,  /* I/O address, IRQ */
	0, 0, 0, NULL, ih_probe };

static int io = 0xd80;
static int irq = 12;
static int mem = 0xcc000;
static char *devname = NULL;

int init_module(void)
{
	int result;

	/* Copy the parameters from insmod into the device structure. */
	ih_device.base_addr = io;
	ih_device.irq       = irq;
	ih_device.mem_start = mem;
	if(devname)
		ih_device.name = devname;

	if ((result = register_netdev(&ih_device)) != 0)
		return result;

	return 0;
}

void
cleanup_module(void)
{
	struct priv *p;
	int i;

	p = (struct priv *) ih_device.priv;

	ih_init(&ih_device,0);

	unregister_netdev(&ih_device);

	if(ih_device.base_addr)
		release_region(ih_device.base_addr, 8);
	else
		free_irq(ih_device.irq, NULL);

	if (p) {
		for(i=0;i<NUM_CHANNELS;i++) 
			if(p->skb[i]) {
				dev_kfree_skb(p->skb[i],FREE_WRITE);
				printk("isachscx: oops non free p->skb[x] buffer\n");
			}
		kfree_s(ih_device.priv, sizeof(struct priv));
	}
}

#endif /* MODULE */


